@@ -1,3 +1,6 @@
+
+.engine_context <- new.env(parent = emptyenv())
+
 #' A reticulate Engine for Knitr
 #'
 #' This provides a `reticulate` engine for `knitr`, suitable for usage when
@@ -21,8 +24,10 @@
 #'
 #' @export
 eng_python <- function(options) {
+  
+  # check for unsupported knitr options
   options <- eng_python_validate_options(options)
-
+  
   # when 'eval = FALSE', we can just return the source code verbatim
   # (skip any other per-chunk work)
   if (identical(options$eval, FALSE)) {
@@ -59,12 +64,13 @@ eng_python <- function(options) {
     }
   }
 
-  context <- new.env(parent = emptyenv())
-  eng_python_initialize(
-    options,
-    context = context,
-    envir = environment()
-  )
+  # environment tracking the labels assigned to newly-created altair charts
+  .engine_context$altair_ids <- new.env(parent = emptyenv())
+  
+  # a list of pending plots / outputs
+  .engine_context$pending_plots <- stack()
+  
+  eng_python_initialize(options = options, envir = environment())
 
   # helper function for extracting range of code, dropping blank lines
   extract <- function(code, range) {
@@ -93,7 +99,7 @@ eng_python <- function(options) {
 
   # iterate over top-level nodes and extract line numbers
   lines <- vapply(parsed$body, function(node) {
-    if(py_has_attr(node, 'decorator_list') && length(node$decorator_list)) {
+    if (py_has_attr(node, "decorator_list") && length(node$decorator_list)) {
       node$decorator_list[[1]]$lineno
     } else {
       node$lineno
@@ -120,7 +126,10 @@ eng_python <- function(options) {
   had_error <- FALSE
 
   # actual outputs to be returned to knitr
-  outputs <- list()
+  outputs <- stack()
+  
+  # 'held' outputs, to be appended at the end (for results = "hold")
+  held_outputs <- stack()
 
   # synchronize state R -> Python
   eng_python_synchronize_before()
@@ -139,9 +148,9 @@ eng_python <- function(options) {
     # extract code to be run
     snippet <- extract(code, range)
 
-    # save last value
-    last_value <- py_last_value()
-
+    # clear the last value object (so we can tell if it was updated)
+    py_compile_eval("'__reticulate_placeholder__'")
+    
     # use trailing semicolon to suppress output of return value
     suppress <- grepl(";\\s*$", snippet)
     compile_mode <- if (suppress) "exec" else "single"
@@ -153,28 +162,54 @@ eng_python <- function(options) {
       py_compile_eval(snippet, compile_mode)
 
     # handle matplotlib output
-    captured <- eng_python_matplotlib_handle_output(captured, last_value, i == length(ranges))
+    captured <- eng_python_autoprint(
+      captured = captured,
+      options  = options,
+      autoshow = i == length(ranges)
+    )
 
-    if (length(context$pending_plots) || !identical(captured, "")) {
+    # emit outputs if we have any
+    has_outputs <-
+      !.engine_context$pending_plots$empty() ||
+      !identical(captured, "")
+    
+    if (has_outputs) {
 
       # append pending source to outputs (respecting 'echo' option)
-      if (!identical(options$echo, FALSE)) {
+      if (!identical(options$echo, FALSE) && !identical(options$results, "hold")) {
         extracted <- extract(code, c(pending_source_index, range[2]))
         output <- structure(list(src = extracted), class = "source")
-        outputs[[length(outputs) + 1]] <- output
+        outputs$push(output)
       }
 
       # append captured outputs (respecting 'include' option)
       if (isTRUE(options$include)) {
 
-        # append captured output
-        if (!identical(captured, ""))
-          outputs[[length(outputs) + 1]] <- captured
-
-        # append captured images / figures
-        for (plot in context$pending_plots)
-          outputs[[length(outputs) + 1]] <- plot
-        context$pending_plots <- list()
+        if (identical(options$results, "hold")) {
+          
+          # append captured output
+          if (!identical(captured, ""))
+            held_outputs$push(captured)
+          
+          # append captured images / figures
+          plots <- .engine_context$pending_plots$data()
+          for (plot in plots)
+            held_outputs$push(plot)
+          .engine_context$pending_plots$clear()
+          
+        } else {
+          
+          # append captured output
+          if (!identical(captured, ""))
+            outputs$push(captured)
+          
+          # append captured images / figures
+          plots <- .engine_context$pending_plots$data()
+          for (plot in plots)
+            outputs$push(plot)
+          .engine_context$pending_plots$clear()
+          
+        }
 
       }
 
@@ -191,32 +226,75 @@ eng_python <- function(options) {
   }
 
   # if we have leftover input, add that now
-  if (!had_error && !identical(options$echo, FALSE) && pending_source_index <= n) {
+  has_leftovers <-
+    !had_error &&
+    !identical(options$echo, FALSE) &&
+    !identical(options$results, "hold") &&
+    pending_source_index <= n
+  
+  if (has_leftovers) {
     leftover <- extract(code, c(pending_source_index, n))
-    outputs[[length(outputs) + 1]] <- structure(
-      list(src = leftover),
-      class = "source"
-    )
+    output <- structure(list(src = leftover), class = "source")
+    outputs$push(output)
+  }
+  
+  # if we were using held outputs, we just inject the source in now
+  if (identical(options$results, "hold")) {
+    output <- structure(list(src = code), class = "source")
+    outputs$push(output)
+  }
+  
+  # if we had held outputs, add those in now (merging text output as appropriate)
+  text_output <- character()
+  
+  held_outputs <- held_outputs$data()
+  for (i in seq_along(held_outputs)) {
+    
+    output <- held_outputs[[i]]
+    if (!is.object(output) && is.character(output)) {
+      
+      # merge text output and save for later
+      text_output <- c(text_output, held_outputs[[i]])
+      
+    } else {
+      
+      # add in pending text output
+      if (length(text_output)) {
+        output <- paste(text_output, collapse = "")
+        outputs$push(output)
+        text_output <- character()
+      }
+      
+      # add in this piece of output
+      outputs$push(held_outputs[[i]])
+    }
+    
+  }
+  
+  # if we have any leftover held output, add in now
+  if (length(text_output)) {
+    output <- paste(text_output, collapse = "")
+    outputs$push(output)
   }
 
   eng_python_synchronize_after()
 
   wrap <- getOption("reticulate.engine.wrap", eng_python_wrap)
-  wrap(outputs, options)
+  wrap(outputs$data(), options)
 
 }
 
-eng_python_initialize <- function(options, context, envir) {
+eng_python_initialize <- function(options, envir) {
 
   if (is.character(options$engine.path))
     use_python(options$engine.path[[1]])
 
   ensure_python_initialized()
+  eng_python_initialize_hooks(options, envir)
 
-  eng_python_initialize_matplotlib(options, context, envir)
 }
 
-eng_python_matplotlib_show <- function(plt, options) {
+eng_python_knit_figure_path <- function(options, suffix = NULL) {
   
   # we need to work in either base.dir or output.dir, depending
   # on which of the two has been requested by the user. (note
@@ -224,7 +302,7 @@ eng_python_matplotlib_show <- function(plt, options) {
   dir <-
     knitr::opts_knit$get("base.dir") %||%
     knitr::opts_knit$get("output.dir")
-
+  
   # move to the requested directory
   dir.create(dir, recursive = TRUE, showWarnings = FALSE)
   owd <- setwd(dir)
@@ -232,12 +310,26 @@ eng_python_matplotlib_show <- function(plt, options) {
   
   # construct plot path
   plot_counter <- yoink("knitr", "plot_counter")
+  number <- plot_counter()
   path <- knitr::fig_path(
-    suffix = options$dev,
+    suffix  = suffix %||% options$dev,
     options = options,
-    number = plot_counter()
+    number  = number
   )
   
+  # ensure parent path exists
+  dir.create(dirname(path), recursive = TRUE, showWarnings = FALSE)
+  
+  # return path
+  path
+  
+}
+
+eng_python_matplotlib_show <- function(plt, options) {
+  
+  # get figure path
+  path <- eng_python_knit_figure_path(options)
+  
   # save the current figure
   dir.create(dirname(path), recursive = TRUE, showWarnings = FALSE)
   plt$savefig(path, dpi = options$dpi)
@@ -248,16 +340,49 @@ eng_python_matplotlib_show <- function(plt, options) {
 
 }
 
-eng_python_initialize_matplotlib <- function(options, context, envir) {
+eng_python_initialize_hooks <- function(options, envir) {
+  
+  # set up hooks for matplotlib modules
+  matplotlib_modules <- c(
+    "matplotlib.artist",
+    "matplotlib.pyplot",
+    "matplotlib.pylab"
+  )
+  
+  for (module in matplotlib_modules) {
+    py_register_load_hook(module, function(...) {
+      eng_python_initialize_matplotlib(options, envir)
+    })
+  }
+  
+  # set up hooks for plotly modules
+  plotly_modules <- c(
+    "plotly.io",
+    "plotlyjs"
+  )
+  
+  for (module in plotly_modules) {
+    py_register_load_hook(module, function(...) {
+      eng_python_initialize_plotly(options, envir)
+    })
+  }
+  
+}
 
-  if (!py_module_available("matplotlib"))
-    return()
+eng_python_initialize_matplotlib <- function(options, envir) {
 
+  # mark initialization done
+  if (identical(.globals$matplotlib_initialized, TRUE))
+    return(TRUE)
+  
+  .globals$matplotlib_initialized <- TRUE
+  
   # attempt to enforce a non-Qt matplotlib backend. this is especially important
   # with RStudio Desktop as attempting to use a Qt backend will cause issues due
   # to mismatched Qt versions between RStudio and Anaconda environments, and
   # will cause crashes when attempting to generate plots
-  if (is_rstudio_desktop()) {
+  testthat <- Sys.getenv("TESTTHAT", unset = NA)
+  if (is_rstudio_desktop() || identical(testthat, "true")) {
 
     matplotlib <- import("matplotlib", convert = TRUE)
 
@@ -265,10 +390,15 @@ eng_python_initialize_matplotlib <- function(options, context, envir) {
     # need to switch backends; otherwise, we can simply request to use a
     # specific one when the backend is initialized later
     sys <- import("sys", convert = FALSE)
-    if ("matplotlib.backends" %in% names(sys$modules))
+    if ("matplotlib.backends" %in% names(sys$modules)) {
       matplotlib$pyplot$switch_backend("agg")
-    else
-      matplotlib$use("agg", warn = FALSE, force = TRUE)
+    } else {
+      version <- numeric_version(matplotlib$`__version__`)
+      if (version < "3.3.0")
+        matplotlib$use("agg", warn = FALSE, force = TRUE)
+      else
+        matplotlib$use("agg", force = TRUE)
+    }
   }
 
   # double-check that we can load 'pyplot' (this can fail if matplotlib
@@ -276,19 +406,9 @@ eng_python_initialize_matplotlib <- function(options, context, envir) {
   if (!py_module_available("matplotlib.pyplot"))
     return()
 
-  # initialize pending_plots list
-  context$pending_plots <- list()
-
   plt <- import("matplotlib.pyplot", convert = FALSE)
 
-  # rudely steal 'plot_counter' (used by default 'show()' implementation below)
-  # and then reset the counter when we're done
-  plot_counter <- yoink("knitr", "plot_counter")
-  defer(plot_counter(reset = TRUE), envir = envir)
-
-  # save + restore old show hook
-  show <- plt$show
-  defer(plt$show <- show, envir = envir)
+  # override show implementation
   plt$show <- function(...) {
 
     # call hook to generate plot
@@ -296,11 +416,11 @@ eng_python_initialize_matplotlib <- function(options, context, envir) {
     graphic <- hook(plt, options)
 
     # update set of pending plots
-    context$pending_plots[[length(context$pending_plots) + 1]] <<- graphic
+    .engine_context$pending_plots$push(graphic)
 
     # return None to ensure no printing of output here (just inclusion of
     # plot as a side effect)
-    r_to_py(NULL)
+    py_none()
 
   }
 
@@ -309,9 +429,24 @@ eng_python_initialize_matplotlib <- function(options, context, envir) {
 
 }
 
+eng_python_initialize_plotly <- function(options, envir) {
+  
+  # mark initialization done
+  if (identical(.globals$plotly_initialized, TRUE))
+    return(TRUE)
+  
+  .globals$plotly_initialized <- TRUE
+  
+  # override the figure 'show' method to just return the plot object itself
+  # the auto-printer will then handle rendering the image as appropriate
+  io <- import("plotly.io", convert = FALSE)
+  io$show <- function(self, ...) self
+  
+}
+
 # synchronize objects R -> Python
 eng_python_synchronize_before <- function() {
-  py_inject_r(envir = getOption("reticulate.engine.environment"))
+  py_inject_r()
 }
 
 # synchronize objects Python -> R
@@ -337,7 +472,7 @@ eng_python_validate_options <- function(options) {
       options[[option]] <- TRUE
     }
   }
-
+  
   options
 }
 
@@ -352,26 +487,138 @@ eng_python_is_matplotlib_output <- function(value) {
 
 }
 
-eng_python_matplotlib_handle_output <- function(captured, last_value, show) {
+eng_python_is_seaborn_output <- function(value) {
+  inherits(value, "seaborn.axisgrid.Grid")
+}
 
-  value <- py_last_value()
+eng_python_is_plotly_plot <- function(value) {
+  inherits(value, "plotly.basedatatypes.BaseFigure")
+}
+
+eng_python_is_altair_chart <- function(value) {
+  
+  # support different API versions, assuming that the class name
+  # otherwise remains compatible
+  classes <- class(value)
+  pattern <- "^altair\\.vegalite\\.v[[:digit:]]+\\.api\\.Chart$"
+  any(grepl(pattern, classes))
+  
+}
+
+eng_python_altair_chart_id <- function(options, ids) {
+  
+  label <- options$label
+  components <- c(label, "altair-viz")
+  if (exists(label, envir = ids)) {
+    id <- get(label, envir = ids)
+    components <- c(components, id + 1)
+    assign(label, id + 1, envir = ids)
+  } else {
+    assign(label, 1L, envir = ids)
+  }
+  
+  paste(components, collapse = "-")
+  
+}
+
+eng_python_autoprint <- function(captured, options, autoshow) {
 
   # bail if no new value was produced by interpreter
-  builtins <- import_builtins(convert = TRUE)
-  if (builtins$id(last_value) == builtins$id(value))
+  value <- py_last_value()
+  if (py_is_none(value))
     return(captured)
+  
+  # ignore placeholder outputs
+  if (inherits(value, "python.builtin.str")) {
+    contents <- py_to_r(value)
+    if (identical(contents, "__reticulate_placeholder__"))
+      return(captured)
+  }
+  
+  # check if output format is html
+  isHtml <- knitr::is_html_output()
 
-  # bail if this isn't matplotlib output
-  if (!eng_python_is_matplotlib_output(value))
-    return(captured)
-
-  # show plot if requested
-  if (show) {
+  if (eng_python_is_matplotlib_output(value) && autoshow) {
+    
+    # handle matplotlib output. note that the default hook installed by
+    # reticulate will update the 'pending_plots' item
     plt <- import("matplotlib.pyplot", convert = TRUE)
     plt$show()
+    return("")
+    
+  } else if (eng_python_is_seaborn_output(value)) {
+    
+    # get figure path
+    path <- eng_python_knit_figure_path(options)
+    value$savefig(path)
+    .engine_context$pending_plots$push(knitr::include_graphics(path))
+    return("")
+    
+  } else if (inherits(value, "pandas.core.frame.DataFrame")) {
+    
+    return(captured)
+    
+  } else if (isHtml && py_has_attr(value, "_repr_html_")) {
+    
+    data <- as_r_value(value$`_repr_html_`())
+    .engine_context$pending_plots$push(knitr::raw_html(data))
+    return("")
+    
+  } else if (eng_python_is_plotly_plot(value) &&
+             py_module_available("psutil") &&
+             py_module_available("kaleido")) {
+    
+    path <- eng_python_knit_figure_path(options)
+    value$write_image(
+      file   = path,
+      width  = options$out.width.px,
+      height = options$out.height.px
+    )
+    .engine_context$pending_plots$push(knitr::include_graphics(path))
+    return("")
+    
+  } else if (eng_python_is_altair_chart(value)) {
+    
+    # set width if it's not already set
+    width <- value$width
+    if (inherits(width, "altair.utils.schemapi.UndefinedType")) {
+      width <- options$altair.fig.width %||% options$out.width.px %||% 810L
+      value <- value$properties(width = width)
+    }
+    
+    # set height if it's not already set
+    height <- value$height
+    if (inherits(height, "altair.utils.schemapi.UndefinedType")) {
+      height <- options$altair.fig.height %||% options$out.height.px %||% 400L
+      value <- value$properties(height = height)
+    }
+    
+    # set a unique id (used for div container for figure)
+    id <- eng_python_altair_chart_id(options, .engine_context$altair_ids)
+    
+    # convert to HTML or PNG as appropriate
+    if (isHtml) {
+      data <- as_r_value(value$to_html(output_div = id))
+      .engine_context$pending_plots$push(knitr::raw_html(data))
+    } else {
+      path <- eng_python_knit_figure_path(options)
+      value$save(path)
+      .engine_context$pending_plots$push(knitr::include_graphics(path))
+    }
+    
+    return("")
+    
+  } else if (py_has_attr(value, "to_html")) {
+    
+    data <- as_r_value(value$to_html())
+    .engine_context$pending_plots$push(knitr::raw_html(data))
+    return("")
+    
+  } else {
+    
+    # nothing special to do
+    return(captured)
+    
   }
 
-  # suppress textual output
-  ""
-
 }
