@@ -25,7 +25,6 @@ NULL
 .globals$class_filters <- list()
 .globals$py_repl_active <- FALSE
 
-
 is_python_initialized <- function() {
   !is.null(.globals$py_config)
 }
@@ -34,6 +33,7 @@ is_python_initialized <- function() {
 ensure_python_initialized <- function(required_module = NULL) {
 
   if (!is_python_initialized()) {
+
     # give delay load modules priority
     use_environment <- NULL
     if (!is.null(.globals$delay_load_module)) {
@@ -46,29 +46,32 @@ ensure_python_initialized <- function(required_module = NULL) {
 
     .globals$py_config <- initialize_python(required_module, use_environment)
 
+    # register interrupt handler
+    signals <- import("rpytools.signals", convert = TRUE)
+    signals$initialize(py_interrupt_handler)
+    
     # remap output streams to R output handlers
     remap_output_streams()
-    
+
     # generate 'R' helper object
-    py_inject_r(envir = globalenv())
-    
+    py_inject_r()
+
     # inject hooks
     py_inject_hooks()
-    
+
     # install required packages
     configure_environment()
 
+    # notify front-end (if any) that Python has been initialized
+    callback <- getOption("reticulate.initialized")
+    if (is.function(callback))
+      callback()
+
   }
 }
 
 initialize_python <- function(required_module = NULL, use_environment = NULL) {
 
-  # disallow initialization of Python within .onLoad()
-  
-  # NOTE: disabled for this release of reticulate as there are too many packages
-  # already forcing initialization of Python on load
-  # check_forbidden_initialization()
-  
   # provide hint to install Miniconda if no Python is found
   python_not_found <- function(msg) {
     hint <- "Use reticulate::install_miniconda() if you'd like to install a Miniconda Python environment."
@@ -82,6 +85,9 @@ initialize_python <- function(required_module = NULL, use_environment = NULL) {
   # find configuration
   config <- py_discover_config(required_module, use_environment)
 
+  # check if R is embedded in an python environment
+  py_embedded <- !is.null(main_process_python_info())
+
   # check for basic python prerequsities
   if (is.null(config)) {
     python_not_found("Installation of Python not found, Python bindings not loaded.")
@@ -94,25 +100,60 @@ initialize_python <- function(required_module = NULL, use_environment = NULL) {
   }
 
   # check numpy version and provide a load error message if we don't satisfy it
-  if (is.null(config$numpy) || config$numpy$version < "1.6")
-    numpy_load_error <- "installation of Numpy >= 1.6 not found"
-  else
-    numpy_load_error <- ""
+  numpy_load_error <- tryCatch(
+
+    expr = {
+      if (is.null(config$numpy) || config$numpy$version < "1.6")
+        "installation of Numpy >= 1.6 not found"
+      else
+        ""
+    },
+
+    error = function(e) "<unknown>"
+
+  )
 
   # if we're a virtual environment then set VIRTUAL_ENV (need to
   # set this before initializing Python so that module paths are
   # set as appropriate)
   if (nzchar(config$virtualenv))
     Sys.setenv(VIRTUAL_ENV = config$virtualenv)
-  
+
   # set R_SESSION_INITIALIZED flag (used by rpy2)
   curr_session_env <- Sys.getenv("R_SESSION_INITIALIZED", unset = NA)
   Sys.setenv(R_SESSION_INITIALIZED = sprintf('PID=%s:NAME="reticulate"', Sys.getpid()))
 
-  # initialize python
+  # prefer utf-8 encoding on Windows in RStudio
+  if (is_rstudio()) {
+    encoding <- Sys.getenv("PYTHONIOENCODING", unset = NA)
+    if (is.na(encoding))
+      Sys.setenv(PYTHONIOENCODING = "utf-8")
+  }
+
+  # munge PATH for python (needed so libraries can be found in some cases)
   oldpath <- python_munge_path(config$python)
-  tryCatch(
-    {
+
+  # initialize python
+  tryCatch({
+
+    # set PYTHONPATH (required to load virtual environments in some cases?)
+    oldpythonpath <- Sys.getenv("PYTHONPATH")
+    newpythonpath <- Sys.getenv(
+      "RETICULATE_PYTHONPATH",
+      unset = paste(
+        config$pythonpath,
+        system.file("python", package = "reticulate"),
+        sep = .Platform$path.sep
+      )
+    )
+
+    local({
+
+      # set PYTHONPATH while we initialize
+      Sys.setenv(PYTHONPATH = newpythonpath)
+      on.exit(Sys.setenv(PYTHONPATH = oldpythonpath), add = TRUE)
+
+      # initialize Python
       py_initialize(config$python,
                     config$libpython,
                     config$pythonhome,
@@ -120,7 +161,11 @@ initialize_python <- function(required_module = NULL, use_environment = NULL) {
                     config$version >= "3.0",
                     interactive(),
                     numpy_load_error)
+
+    })
+
     },
+
     error = function(e) {
       Sys.setenv(PATH = oldpath)
       if (is.na(curr_session_env)) {
@@ -130,19 +175,30 @@ initialize_python <- function(required_module = NULL, use_environment = NULL) {
       }
       stop(e)
     }
+
   )
 
   # set available flag indicating we have py bindings
   config$available <- TRUE
 
-  # add our python scripts to the search path
-  py_run_string_impl(paste0("import sys; sys.path.append('",
-                       system.file("python", package = "reticulate") ,
-                       "')"))
+  if (py_embedded) {
+    
+    # we need to insert path to rpytools directly for embedded R
+    path <- system.file("python", package = "reticulate")
+    fmt <- "import sys; sys.path.append(%s)"
+    cmd <- sprintf(fmt, shQuote(path))
+    
+    py_run_string_impl(cmd)
+    
+  }
 
   # ensure modules can be imported from the current working directory
   py_run_string_impl("import sys; sys.path.insert(0, '')")
 
+  # if this is a conda installation, set QT_QPA_PLATFORM_PLUGIN_PATH
+  # https://github.com/rstudio/reticulate/issues/586
+  py_set_qt_qpa_platform_plugin_path(config)
+
   # notify the user if the loaded version of Python isn't the same
   # as the requested version of python
   local({
@@ -176,61 +232,115 @@ initialize_python <- function(required_module = NULL, use_environment = NULL) {
     }
 
   })
-  
-  # notify front-end (if any) that Python has been initialized
-  callback <- getOption("reticulate.initialized")
-  if (is.function(callback))
-    callback()
 
   # return config
   config
 }
 
 check_forbidden_initialization <- function() {
-  
+
   if (is_python_initialized())
     return(FALSE)
-  
+
   override <- getOption(
     "reticulate.allow.package.initialization",
     default = FALSE
   )
-  
+
   if (identical(override, TRUE))
     return(FALSE)
-  
+
   calls <- sys.calls()
   frames <- sys.frames()
-  
+
   for (i in seq_along(calls)) {
-  
+
     call <- calls[[i]]
     frame <- frames[[i]]
     if (!identical(call[[1]], as.name("runHook")))
       next
-    
+
     bad <-
       identical(call[[2]], ".onLoad") ||
       identical(call[[2]], ".onAttach")
-    
+
     if (!bad)
       next
-    
+
     pkgname <- tryCatch(
         get("pkgname", envir = frame),
         error = function(e) "<unknown>"
       )
-      
+
     fmt <- paste(
       "package '%s' attempted to initialize Python in %s().",
       "Packages should not initialize Python themselves; rather, Python should",
       "be loaded on-demand as requested by the user of the package. Please see",
       "vignette(\"python_dependencies\", package = \"reticulate\") for more details."
     )
-    
+
     msg <- sprintf(fmt, pkgname, call[[2]])
     warning(msg)
-    
+
   }
-  
+
+}
+
+check_forbidden_install <- function(label) {
+
+  # escape hatch for users who know, or claim to know, what they're doing
+  envvar <- Sys.getenv("_RETICULATE_I_KNOW_WHAT_IM_DOING_", unset = NA)
+  if (identical(tolower(envvar), "true"))
+    return(FALSE)
+
+  # if this is being called as part of R CMD check, then warn
+  # (error in future versions)
+  if (is_r_cmd_check()) {
+    fmt <- "cannot install %s during R CMD check"
+    msg <- sprintf(fmt, label)
+    warning(msg)
+    return(TRUE)
+  }
+
+  FALSE
+
+}
+
+py_set_qt_qpa_platform_plugin_path <- function(config) {
+
+  # only done on Windows since that's where we see all the issues
+  if (!is_windows())
+    return(FALSE)
+
+  # get python homes (note that multiple homes may be specified)
+  homes <- strsplit(config$pythonhome, ";", fixed = TRUE)[[1]]
+  for (home in homes) {
+
+    # build some candidate paths to the plugins directory
+    candidates <- c(
+      file.path(home, "Library/plugins/platforms"),
+      file.path(home, "../../Library/plugins/platforms")
+    )
+
+    # check and see if any of these exist
+    paths <- candidates[file.exists(candidates)]
+    if (length(paths) == 0)
+      next
+
+    # we found a path; use it
+    path <- normalizePath(paths[[1]], winslash = "/", mustWork = TRUE)
+    path <- gsub("/", "\\\\", path, fixed = TRUE)
+
+    fmt <- "import os; os.environ['QT_QPA_PLATFORM_PLUGIN_PATH'] = '%s'"
+    cmd <- sprintf(fmt, path)
+    py_run_string_impl(cmd)
+
+    # return TRUE to indicate success
+    return(TRUE)
+
+  }
+
+  # failed to find folder; nothing to do
+  FALSE
+
 }
