@@ -32,13 +32,17 @@
 #' @param quiet Boolean; print a startup banner when launching the REPL? If
 #'   `TRUE`, the banner will be suppressed.
 #'
+#' @param input Python code to be run within the REPL. Setting this can be
+#'   useful if you'd like to drive the Python REPL programmatically.
+#'
 #' @seealso [py], for accessing objects created using the Python REPL.
 #'
 #' @importFrom utils packageVersion
 #' @export
 repl_python <- function(
-  module = NULL,
-  quiet = getOption("reticulate.repl.quiet", default = FALSE))
+  module  = NULL,
+  quiet   = getOption("reticulate.repl.quiet", default = FALSE),
+  input   = NULL)
 {
   # load module if requested
   if (is.character(module))
@@ -54,6 +58,10 @@ repl_python <- function(
   if (is.function(teardown)) {
     on.exit(teardown(), add = TRUE)
   }
+  
+  # split provided code on newlines
+  if (!is.null(input))
+    input <- unlist(strsplit(input, "\n", fixed = TRUE))
 
   # import other required modules for the REPL
   sys <- import("sys", convert = TRUE)
@@ -94,7 +102,7 @@ repl_python <- function(
   #
   # we return an environment of functions bound in a local environment
   # so that hook can manipulate the buffer if required
-  buffer <- new_stack()
+  buffer <- stack(mode = "character")
 
   # command compiler (used to check if we've received a complete piece
   # of Python input)
@@ -116,6 +124,12 @@ repl_python <- function(
     }
     failed
   }
+  
+  handle_interrupt <- function(condition) {
+    # swallow interrupts -- don't allow interrupted Python code to
+    # exit the REPL; we should only exit when an interrupt is sent
+    # when no Python code is executing
+  }
 
   repl <- function() {
     
@@ -129,10 +143,19 @@ repl_python <- function(
         sys$stderr$flush()
       
     }, add = TRUE)
-
-    # read user input
+    
+    # read input (either from user or from code)
     prompt <- if (buffer$empty()) ">>> " else "... "
-    contents <- readline(prompt = prompt)
+    if (is.null(input)) {
+      contents <- readline(prompt = prompt)
+    } else if (length(input)) {
+      contents <- input[[1L]]
+      input <<- tail(input, n = -1L)
+      writeLines(paste(prompt, contents), con = stdout())
+    } else {
+      quit_requested <<- TRUE
+      return()
+    }
 
     # NULL implies the user sent EOF -- time to leave
     if (is.null(contents)) {
@@ -161,6 +184,24 @@ repl_python <- function(
       if (isTRUE(status))
         return()
     }
+    
+    # run hook provided by front-end, to notify that we're now busy
+    hook <- getOption("reticulate.repl.busy")
+    if (is.function(hook)) {
+      
+      # run once now to indicate we're about to run
+      status <- tryCatch(hook(TRUE), error = identity)
+      if (inherits(status, "error"))
+        warning(status)
+      
+      # run again on exit to indicate we're done
+      on.exit({
+        status <- tryCatch(hook(FALSE), error = identity)
+        if (inherits(status, "error"))
+          warning(status)
+      }, add = TRUE)
+      
+    }
 
     # special handling for top-level commands (when buffer is empty)
     if (buffer$empty()) {
@@ -233,7 +274,12 @@ repl_python <- function(
 
       # submit previous code
       pasted <- paste(previous, collapse = "\n")
-      tryCatch(py_compile_eval(pasted, capture = FALSE), error = handle_error)
+      
+      tryCatch(
+        py_compile_eval(pasted, capture = FALSE),
+        error = handle_error,
+        interrupt = handle_interrupt
+      )
 
       # now, handle the newest line of code submitted
       buffer$set(contents)
@@ -248,7 +294,12 @@ repl_python <- function(
     # otherwise, we should have received a code output object
     # so we can just run the code submitted thus far
     buffer$clear()
-    tryCatch(py_compile_eval(code, capture = FALSE), error = handle_error)
+    
+    tryCatch(
+      py_compile_eval(code, capture = FALSE),
+      error = handle_error,
+      interrupt = handle_interrupt
+    )
 
   }
 
@@ -290,6 +341,7 @@ repl_python <- function(
       break
 
     repl()
+    
   }
 
 }
