@@ -25,12 +25,14 @@ public:
   }
 
   PyObject* get() const {
+    
     SEXP pyObject = getFromEnvironment("pyobj");
     if (pyObject != R_NilValue) {
       PyObject* obj = (PyObject*)R_ExternalPtrAddr(pyObject);
       if (obj != NULL)
         return obj;
     }
+    
     Rcpp::stop("Unable to access object (object is from previous session and is now invalid)");
   }
 
@@ -51,13 +53,13 @@ public:
   }
 
   void set(PyObject* object) {
-    SEXP xptr = R_MakeExternalPtr((void*) object, R_NilValue, R_NilValue);
+    Rcpp::RObject xptr = R_MakeExternalPtr((void*) object, R_NilValue, R_NilValue);
     R_RegisterCFinalizer(xptr, python_object_finalize);
     assign("pyobj", xptr);
   }
 
   bool convert() const {
-    SEXP pyObject = getFromEnvironment("convert");
+    Rcpp::RObject pyObject = getFromEnvironment("convert");
     if (pyObject == R_NilValue)
       return true;
     else
